###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""Module with PID aliases - shorthand names for the actual PID variables."""

ALIASES = {
    "Brunel_DLLK": "probe_Brunel_PIDK",
    "Brunel_DLLd": "probe_Brunel_PIDd",
    "Brunel_DLLe": "probe_Brunel_PIDe",
    "Brunel_DLLmu": "probe_Brunel_PIDmu",
    "Brunel_DLLp": "probe_Brunel_PIDp",
    "Brunel_ETA": "probe_Brunel_ETA",
    "Brunel_HasRich": "probe_Brunel_hasRich",
    "Brunel_IPCHI2": "probe_Brunel_IPCHI2",
    "Brunel_InMuonAcc": "probe_Brunel_InMuonAcc",
    "Brunel_IsMuon": "probe_Brunel_isMuon",
    "Brunel_MC12TuneV2_ProbNNe": "probe_Brunel_MC12TuneV2_ProbNNe",
    "Brunel_MC12TuneV2_ProbNNghost": "probe_Brunel_MC12TuneV2_ProbNNghost",
    "Brunel_MC12TuneV2_ProbNNk": "probe_Brunel_MC12TuneV2_ProbNNk",
    "Brunel_MC12TuneV2_ProbNNmu": "probe_Brunel_MC12TuneV2_ProbNNmu",
    "Brunel_MC12TuneV2_ProbNNp": "probe_Brunel_MC12TuneV2_ProbNNp",
    "Brunel_MC12TuneV2_ProbNNpi": "probe_Brunel_MC12TuneV2_ProbNNpi",
    "Brunel_MC12TuneV3_ProbNNe": "probe_Brunel_MC12TuneV3_ProbNNe",
    "Brunel_MC12TuneV3_ProbNNghost": "probe_Brunel_MC12TuneV3_ProbNNghost",
    "Brunel_MC12TuneV3_ProbNNk": "probe_Brunel_MC12TuneV3_ProbNNk",
    "Brunel_MC12TuneV3_ProbNNmu": "probe_Brunel_MC12TuneV3_ProbNNmu",
    "Brunel_MC12TuneV3_ProbNNp": "probe_Brunel_MC12TuneV3_ProbNNp",
    "Brunel_MC12TuneV3_ProbNNpi": "probe_Brunel_MC12TuneV3_ProbNNpi",
    "Brunel_MC12TuneV4_ProbNNe": "probe_Brunel_MC12TuneV4_ProbNNe",
    "Brunel_MC12TuneV4_ProbNNghost": "probe_Brunel_MC12TuneV4_ProbNNghost",
    "Brunel_MC12TuneV4_ProbNNk": "probe_Brunel_MC12TuneV4_ProbNNk",
    "Brunel_MC12TuneV4_ProbNNmu": "probe_Brunel_MC12TuneV4_ProbNNmu",
    "Brunel_MC12TuneV4_ProbNNp": "probe_Brunel_MC12TuneV4_ProbNNp",
    "Brunel_MC12TuneV4_ProbNNpi": "probe_Brunel_MC12TuneV4_ProbNNpi",
    "Brunel_MC15TuneV1_ProbNNe": "probe_Brunel_MC15TuneV1_ProbNNe",
    "Brunel_MC15TuneV1_ProbNNghost": "probe_Brunel_MC15TuneV1_ProbNNghost",
    "Brunel_MC15TuneV1_ProbNNk": "probe_Brunel_MC15TuneV1_ProbNNk",
    "Brunel_MC15TuneV1_ProbNNmu": "probe_Brunel_MC15TuneV1_ProbNNmu",
    "Brunel_MC15TuneV1_ProbNNp": "probe_Brunel_MC15TuneV1_ProbNNp",
    "Brunel_MC15TuneV1_ProbNNpi": "probe_Brunel_MC15TuneV1_ProbNNpi",
    "Brunel_MuonUnbiased": "probe_Brunel_MuonUnbiased",
    "Brunel_P": "probe_Brunel_P",
    "Brunel_PT": "probe_Brunel_PT",
    "Brunel_TRACK_GHOSTPROB": "probe_Brunel_TRACK_GHOSTPROB",
    "Brunel_TRCHI2NDOF": "probe_Brunel_TRCHI2NDOF",
    "DLLK": "probe_PIDK",
    "DLLd": "probe_PIDd",
    "DLLe": "probe_PIDe",
    "DLLmu": "probe_PIDmu",
    "DLLp": "probe_PIDp",
    "ETA": "probe_ETA",
    "HasRich": "probe_hasRich",
    "IPCHI2": "probe_IPCHI2",
    "InMuonAcc": "probe_InMuonAcc",
    "IsMuon": "probe_isMuon",
    "MC12TuneV2_ProbNNe": "probe_MC12TuneV2_ProbNNe",
    "MC12TuneV2_ProbNNghost": "probe_MC12TuneV2_ProbNNghost",
    "MC12TuneV2_ProbNNk": "probe_MC12TuneV2_ProbNNk",
    "MC12TuneV2_ProbNNmu": "probe_MC12TuneV2_ProbNNmu",
    "MC12TuneV2_ProbNNp": "probe_MC12TuneV2_ProbNNp",
    "MC12TuneV2_ProbNNpi": "probe_MC12TuneV2_ProbNNpi",
    "MC12TuneV3_ProbNNe": "probe_MC12TuneV3_ProbNNe",
    "MC12TuneV3_ProbNNghost": "probe_MC12TuneV3_ProbNNghost",
    "MC12TuneV3_ProbNNk": "probe_MC12TuneV3_ProbNNk",
    "MC12TuneV3_ProbNNmu": "probe_MC12TuneV3_ProbNNmu",
    "MC12TuneV3_ProbNNp": "probe_MC12TuneV3_ProbNNp",
    "MC12TuneV3_ProbNNpi": "probe_MC12TuneV3_ProbNNpi",
    "MC12TuneV4_ProbNNe": "probe_MC12TuneV4_ProbNNe",
    "MC12TuneV4_ProbNNghost": "probe_MC12TuneV4_ProbNNghost",
    "MC12TuneV4_ProbNNk": "probe_MC12TuneV4_ProbNNk",
    "MC12TuneV4_ProbNNmu": "probe_MC12TuneV4_ProbNNmu",
    "MC12TuneV4_ProbNNp": "probe_MC12TuneV4_ProbNNp",
    "MC12TuneV4_ProbNNpi": "probe_MC12TuneV4_ProbNNpi",
    "MC15TuneV1_ProbNNe": "probe_MC15TuneV1_ProbNNe",
    "MC15TuneV1_ProbNNghost": "probe_MC15TuneV1_ProbNNghost",
    "MC15TuneV1_ProbNNk": "probe_MC15TuneV1_ProbNNk",
    "MC15TuneV1_ProbNNmu": "probe_MC15TuneV1_ProbNNmu",
    "MC15TuneV1_ProbNNp": "probe_MC15TuneV1_ProbNNp",
    "MC15TuneV1_ProbNNpi": "probe_MC15TuneV1_ProbNNpi",
    "MuonUnbiased": "probe_MuonUnbiased",
    "P": "probe_P",
    "PT": "probe_PT",
    "TRACK_GHOSTPROB": "probe_TRACK_GHOSTPROB",
    "TRCHI2NDOF": "probe_TRCHI2NDOF",
    "nSPDhits": "nSPDhits",
    "nSPDhits_Brunel": "nSPDhits_Brunel",
    "nTracks": "nTracks",
    "nTracks_Brunel": "nTracks_Brunel",
    "trackcharge": "probe_trackcharge",
}
